import torch
import torch.nn.functional as tf
from scipy.signal import check_COLA, get_window


class STFT(torch.nn.Module):
    def __init__(self, win_len=1024, win_hop=512, fft_len=1024,
                 enframe_mode='continue', win_type='hann',
                 win_sqrt=False, pad_center=True):
        """
        Implement of STFT using 1D convolution and 1D transpose convolutions.
        Implement of framing the signal in 2 ways, `break` and `continue`.
        `break` method is a kaldi-like framing.
        `continue` method is a librosa-like framing.

        Args:
            win_len (int): Number of points in one frame.  Defaults to 1024.
            win_hop (int): Number of framing stride. Defaults to 512.
            fft_len (int): Number of DFT points. Defaults to 1024.
            enframe_mode (str, optional): `break` and `continue`.
            Defaults to 'continue'.
            win_type (str, optional): The type of window to create.
            Defaults to 'hann'.
            win_sqrt (bool, optional): using square root window.
            Defaults to True.
            pad_center (bool, optional): `perfect reconstruction` opts.
            Defaults to True.
        """
        super(STFT, self).__init__()
        assert enframe_mode in ['break', 'continue']
        assert fft_len >= win_len
        self.win_len = win_len
        self.win_hop = win_hop
        self.fft_len = fft_len
        self.mode = enframe_mode
        self.win_type = win_type
        self.win_sqrt = win_sqrt
        self.pad_center = pad_center
        self.pad_amount = self.fft_len // 2

        en_k, fft_k, ifft_k, ola_k = self.__init_kernel__()
        self.register_buffer('en_k', en_k)
        self.register_buffer('fft_k', fft_k)
        self.register_buffer('ifft_k', ifft_k)
        self.register_buffer('ola_k', ola_k)

    def __init_kernel__(self):
        """
        Generate enframe_kernel, fft_kernel, ifft_kernel
        and overlap-add kernel.
        ** enframe_kernel: Using conv1d layer and identity matrix.
        ** fft_kernel: Using linear layer for matrix multiplication. In fact,
        enframe_kernel and fft_kernel can be combined, But for the sake of
        readability, I took the two apart.
        ** ifft_kernel, pinv of fft_kernel.
        ** overlap-add kernel, just like enframe_kernel, but transposed.

        Returns:
            tuple: four kernels.
        """
        enframed_kernel = torch.eye(self.fft_len)[:, None, :]
        fft_kernel = torch.rfft(torch.eye(self.fft_len), 1)
        if self.mode == 'break':
            enframed_kernel = torch.eye(self.win_len)[:, None, :]
            fft_kernel = fft_kernel[:self.win_len]
        fft_kernel = torch.cat(
            (fft_kernel[:, :, 0], fft_kernel[:, :, 1]), dim=1)
        ifft_kernel = torch.pinverse(fft_kernel)[:, None, :]
        window = get_window(self.win_type, self.win_len)

        self.perfect_reconstruct = check_COLA(
            window,
            self.win_len,
            self.win_len-self.win_hop)
        window = torch.FloatTensor(window)
        if self.mode == 'continue':
            left_pad = (self.fft_len - self.win_len)//2
            right_pad = left_pad + (self.fft_len - self.win_len) % 2
            window = tf.pad(window, (left_pad, right_pad))
        if self.win_sqrt:
            self.padded_window = window
            window = torch.sqrt(window)
        else:
            self.padded_window = window**2

        fft_kernel = fft_kernel.T * window
        ifft_kernel = ifft_kernel * window
        ola_kernel = torch.eye(self.fft_len)[:self.win_len, None, :]
        if self.mode == 'continue':
            ola_kernel = torch.eye(self.fft_len)[:, None, :self.fft_len]
        return enframed_kernel, fft_kernel, ifft_kernel, ola_kernel

    def is_perfect(self):
        """
        Whether the parameters win_len, win_hop and win_sqrt
        obey constants overlap-add(COLA)

        Returns:
            bool: Return true if parameters obey COLA.
        """
        return self.perfect_reconstruct and self.pad_center

    def transform(self, inputs, return_type='magphase'):
        """Take input data (audio) to STFT domain.

        Args:
            inputs (tensor): Tensor of floats,
            with shape (num_batch, num_samples)
            return_type (str, optional): return (mag, phase) when `magphase`,
            return (real, imag) when `realimag`. Defaults to 'magphase'.

        Returns:
            tuple: (mag, phase) when `magphase`, return (real, imag) when
            `realimag`. Defaults to 'magphase', each elements with shape
            [num_batch, num_frequencies, num_frames]
        """
        assert return_type in ['magphase', 'realimag']
        if inputs.dim() == 2:
            inputs = torch.unsqueeze(inputs, 1)
        self.num_samples = inputs.size(-1)
        if self.pad_center:
            inputs = tf.pad(
                inputs, (self.pad_amount, self.pad_amount), mode='reflect')
        enframe_inputs = tf.conv1d(inputs, self.en_k, stride=self.win_hop)
        outputs = torch.transpose(enframe_inputs, 1, 2)
        outputs = tf.linear(outputs, self.fft_k)
        outputs = torch.transpose(outputs, 1, 2)
        dim = self.fft_len//2+1
        real = outputs[:, :dim, :]
        imag = outputs[:, dim:, :]
        if return_type == 'realimag':
            return real, imag
        else:
            mags = torch.sqrt(real**2+imag**2)
            phase = torch.atan2(imag, real)
            return mags, phase

    def inverse(self, input1, input2, input_type='magphase'):
        """Call the inverse STFT (iSTFT), given tensors produced
        by the `transform` function.

        Args:
            input1 (tensors): Magnitude/Real-part of STFT with shape
            [num_batch, num_frequencies, num_frames]
            input2 (tensors): Phase/Imag-part of STFT with shape [
            [num_batch, num_frequencies, num_frames]
            input_type (str, optional): Mathematical meaning of input tensor's.
            Defaults to 'magphase'.

        Returns:
            tensors: Reconstructed audio given magnitude and phase. Of
                shape [num_batch, num_samples]
        """
        assert input_type in ['magphase', 'realimag']
        if input_type == 'realimag':
            real, imag = input1, input2
        else:
            real = input1*torch.cos(input2)
            imag = input1*torch.sin(input2)
        inputs = torch.cat([real, imag], dim=1)
        outputs = tf.conv_transpose1d(inputs, self.ifft_k, stride=self.win_hop)
        t = (self.padded_window[None, :, None]).repeat(1, 1, inputs.size(-1))
        t = t.to(inputs.device)
        coff = tf.conv_transpose1d(t, self.ola_k, stride=self.win_hop)
        rm_start, rm_end = self.pad_amount, self.pad_amount+self.num_samples
        outputs = outputs[..., rm_start:rm_end]
        coff = coff[..., rm_start:rm_end]
        coffidx = torch.where(coff > 1e-8)
        outputs[coffidx] = outputs[coffidx]/(coff[coffidx])
        return outputs.squeeze(dim=1)

    def forward(self, inputs):
        """Take input data (audio) to STFT domain and then back to audio.

        Args:
            inputs (tensor): Tensor of floats,
            with shape [num_batch, num_samples]

        Returns:
            tensor: Reconstructed audio given magnitude and phase.
            Of shape [num_batch, num_samples]
        """
        mag, phase = self.transform(inputs)
        rec_wav = self.inverse(mag, phase)
        return rec_wav



import torch as th
import numpy as np
from .torch_fbank import FBANK
import torch.nn.functional as tf


class MFCC(th.nn.Module):
    def __init__(self, win_len, win_hop, fft_len, sr,
                 win_type='hann', power=2.0, n_mel=128, n_mfcc=13,
                 dct_type=2, norm='ortho', ref=1.0,
                 amin=1e-10, top_db=80.0,
                 center=True, enframe_mode='continue'):
        """
        Implement of MFCC using 1D convolution and 1D transpose convolutions.

        Args:
            win_len (int): See `conv_stft.STFT` for details.
            win_hop (int): As above
            fft_len (int): As above
            sr (int): Sample rate of signal
            win_type (str, optional): see `conv_stft.STFT` for details.
            Defaults to 'hann'.
            power (float, optional): Do power to spectrum. Defaults to 2.0.
            n_mel (int, optional): Number of mel filter banks. Defaults to 128.
            n_mfcc (int, optional): Mel Cepstral Coefficient. Defaults to 13.
            dct_type (int, optional): Type of DCT transform. Defaults to 2.
            norm (str, optional): Do normalization. Defaults to 'ortho'.
            ref (float, optional): See `librosa.core.power_to_db` for details.
            amin (float, optional): As above. Defaults to 1e-17.
            top_db (float, optional): As above. Defaults to 80.0.
            center (bool, optional): If padding the input signal.
            Defaults to True.
            enframe_mode (str, optional): default enframe method in librosa.
        """
        super(MFCC, self).__init__()
        self.fbank = FBANK(
            win_len, win_hop, fft_len, sr,
            win_type, power, n_mel, ref, amin, top_db,
            center=center, enframe_mode=enframe_mode)
        self.n_mel = n_mel
        self.n_mfcc = n_mfcc
        self.dct_type = dct_type
        self.norm = norm
        dct_k = self.__init_kernel__()
        self.register_buffer('dct_k', dct_k.T)

    def __init_kernel__(self):
        """
        Return the Discrete Cosine Transform of arbitrary type sequence x.
        The details of four kind of DCT can be viewed here:
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.fftpack.dct.html
        Type 1 is not supported, for memory considerations.

        Returns:
            tensors: The DCt transform matrix
        """
        n = th.arange(self.n_mel)[:, None].float()
        k = th.arange(self.n_mfcc)[:, None].float()
        dct_kernel = None
        if self.dct_type == 2:
            dct_kernel = 2.0 * \
                th.cos(np.pi*th.matmul(2.*n+1.0, k.T)/2./self.n_mel)
            if self.norm == 'ortho':
                dct_kernel[:, 0] *= np.sqrt(1/self.n_mel/4.)
                dct_kernel[:, 1:] *= np.sqrt(1/self.n_mel/2.)
        elif self.dct_type == 3:
            dct_kernel = 2.0*th.cos(np.pi*th.matmul(n, 2*k.T+1)/(2*self.n_mel))
            dct_kernel[0, :] = 0.
            if self.norm == 'ortho':
                dct_kernel *= np.sqrt(1/(2*self.n_mel))
        elif self.dct_type == 4:
            dct_kernel = 2.0 * \
                th.cos(np.pi*th.matmul(2.*n+1.0, 2.*k.T+1)/4./self.n_mel)
            if self.norm == 'ortho':
                dct_kernel *= np.sqrt(1/self.n_mel/2)
        else:
            raise RuntimeError(
                "Type {} is not supported.".format(self.dct_type))
        return dct_kernel

    def forward(self, inputs):
        """
        Take input data (audio) to mfcc feature.

        Args:
            inputs (tensor): Tensor of floats, with shape
            [num_batch, num_samples]

        Returns:
            tensor: mfcc feature of shape [num_batch, num_mfcc]
        """
        fbank_spec = self.fbank(inputs)
        if self.dct_type == 3:  # dct_type == 3
            mfcc_spec = tf.linear(fbank_spec, self.dct_k)
            if self.norm != 'ortho':
                mfcc_spec += fbank_spec[:, :, 0:1]
            else:
                mfcc_spec += fbank_spec[:, :, 0:1]*np.sqrt(1/self.n_mel)
        else:  # dct_type == 2/4
            mfcc_spec = tf.linear(fbank_spec, self.dct_k)
        return mfcc_spec



import torch as th
import torch.nn.functional as tf
import numpy as np
from .conv_stft import STFT
from librosa.filters import mel as filters_mel


class FBANK(th.nn.Module):
    def __init__(self, win_len, win_hop, fft_len, sr,
                 win_type='hann', power=2.0, n_mel=128,
                 ref=1.0, amin=1e-17, top_db=80.0,
                 center=True, enframe_mode='continue'):
        """
        Implement of Fbank using 1D convolution and 1D transpose convolutions.

        Args:
            win_len (int): See `conv_stft.STFT` for details.
            win_hop (int): As above
            fft_len (int): As above
            sr (int): Sample rate of signal
            win_type (str, optional): see `conv_stft.STFT` for details.
            Defaults to 'hann'.
            power (float, optional): Do power to spectrum. Defaults to 2.0.
            n_mel (int, optional): Number of mel filter banks. Defaults to 128.
            ref (float, optional): See `librosa.core.power_to_db` for details.
            amin (float, optional): As above. Defaults to 1e-17.
            top_db (float, optional): As above. Defaults to 80.0.
            center (bool, optional): If padding the input signal.
            Defaults to True.
            enframe_mode (str, optional): default enframe method in librosa.
            Defaults to True.
        """
        super(FBANK, self).__init__()
        self.stft = STFT(
            win_len=win_len,
            win_hop=win_hop,
            fft_len=fft_len,
            pad_center=center,
            win_type=win_type,
            enframe_mode=enframe_mode)
        self.sr = sr
        self.fft_len = fft_len
        self.power = power
        self.n_mel = n_mel
        mel_k = self.__init_kernel__()
        self.ref = ref
        self.amin = amin
        self.top_db = top_db
        self.register_buffer('mel_k', mel_k)

    def power2db(self, spec):
        """
        See `librosa.core.power_to_db`. Function as dB-scale.

        Args:
            spec (tensors): Input fbank Feature.

        Returns:
            tensors: db-scale of input feature.
        """
        assert self.amin > 0
        amin = th.ones_like(spec)*self.amin
        ref_value = th.ones_like(spec)*np.abs(self.ref)
        log_spec = 10.0*th.log10(th.max(amin, spec))
        log_spec -= 10.0*th.log10(th.max(amin, ref_value))
        if self.top_db is not None:
            assert self.top_db > 0
            peak_th = th.ones_like(log_spec)*(th.max(log_spec)-self.top_db)
            log_spec = th.max(log_spec, peak_th)
        return log_spec

    def __init_kernel__(self):
        """
        Create a Filterbank matrix to combine FFT bins into Mel-frequency bins

        Returns:
            tensors: A Filterbank matrix
        """
        mel_kernel = th.FloatTensor(filters_mel(
            self.sr, self.fft_len, n_mels=self.n_mel))
        return mel_kernel

    def forward(self, inputs):
        """
        Take input data (audio) to fbank feature.

        Args:
            inputs (tensor): Tensor of floats, with shape
            [num_batch, num_samples]

        Returns:
            tensor: fbank feature of shape [num_batch, num_mels]
        """
        real, imag = self.stft.transform(inputs, return_type='realimag')
        amp_feature = th.sqrt((real)**2 + (imag)**2)
        amp_feature_power = th.pow(amp_feature, self.power)
        mel_spec = tf.linear(amp_feature_power.transpose(1, 2), self.mel_k)
        mel_spec = self.power2db(mel_spec)
        return mel_spec



from setuptools import find_packages, setup

NAME = 'torch_mfcc'
VERSION = "0.1.3"
REQUIREMENTS = [
    'numpy',
    'scipy',
    'librosa'
]

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name=NAME,
      version=VERSION,
      description="A librosa's STFT/FBANK/MFCC implement based on Torch",
      long_description=long_description,
      long_description_content_type="text/markdown",
      url="https://github.com/echocatzh/torch-mfcc",
      author="Shimin Zhang",
      author_email="shmzhang@npu-aslp.org",
      packages=["torch_mfcc"],
      install_requires=REQUIREMENTS,
    #   python_requires=">=3.5",
      license="MIT")